#ifdef HAVE_CONFIG_H
#include <config.h>
#else
#define VERSION "0.0.0-missing"
#endif

#include <assert.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>



/* convert signal name to signal number */
static int
signum(const char *namer)
{
	const char *name;
	assert(namer != NULL);
	if (strncmp(namer, "SIG", 3) == 0) {
		name = namer + 3;
	} else {
		name = namer;
	}

	if (strcmp(name, "USR1") == 0) {
		return SIGUSR1;
	} else if (strcmp(name, "USR2") == 0) {
		return SIGUSR2;
	} else if (strcmp(name, "HUP") == 0) {
		return SIGHUP;
	} else if (strcmp(name, "INT") == 0) {
		return SIGINT;
	} else if (strcmp(name, "KILL") == 0) {
		return SIGKILL;
	} else if (strcmp(name, "TERM") == 0) {
		return SIGTERM;
	} else if (strcmp(name, "ABRT") == 0) {
		return SIGABRT;
	} else if (strcmp(name, "SEGV") == 0) {
		return SIGSEGV;
	} else if (strcmp(name, "TRAP") == 0) {
		return SIGTRAP;
	} else {
		int n = atoi(name);
		fprintf(stderr, "signal %s = %d\n", name, n);
		return n;
	}
}



static void
ok_sh(int sig)
{
	exit(EXIT_SUCCESS);
}



static void
nok_sh(int sig)
{
	exit(EXIT_FAILURE);
}



static void
print_child_cmdline(int argc, char **argv, int skip)
{
	int i;
	fprintf(stderr, "%s", argv[skip]);
	for (i = skip + 1; i < argc; i++) {
		fprintf(stderr, " %s", argv[i]);
	}
}



static void
print_usage(char *self, int full)
{
	fprintf(stderr, "usage: %s [OPTIONS] CHILD [CHILD OPTON ...]\n",
			self);
	if (full == 1) {
		fprintf(stderr, "\
\n\
options:\n\
  -t N  Set child timeout to N seconds\n\
  -o S  Use S as OK-signal from child\n\
  -r S  Replace child's signal S with SIG_IGN\n\
  -n S  Use S as NOK-signal if starting CHILD fails\n\
  -N    Don't use NOK-signal (will sleep for timeout seconds if CHILD fails)\n\
  -k S  Use signal S to kill CHIKD if CHILD times out (defaults to SIGTERM)\n\
  -K    Don't kill CHILD even if CHILD times out\n\
  -u    Use CHILD's return code if CHILD returns before timeout)\n\
  -v    Be verbose\n\
  -h    Print this help and exit\n\
  -V    Print version information and exit\n\
  --    Stop parsing options for %s\n",
				self);
	}
}



static void
print_version(void)
{
	printf("sigwrapper v%s\n", VERSION);
}



int
main(int argc, char **argv)
{
	int ok_sig = SIGUSR1;
	int nok_sig = SIGUSR2;
	int rpl_sig = SIGUSR1;
	int kill_sig = SIGTERM;
	int timeout = 10;
	int use_ec = 0;
	int use_alarm = 1;
	int verbose = 0;

	pid_t pid;

	if (argc == 1) {
		print_usage(argv[0], 0);
		return EXIT_SUCCESS;
	}

	while (1) {
		int c;

		c = getopt(argc, argv, "t:o:r:n:Nk:KuhvV");
		if (c == -1) {
			break;
		}

		switch (c) {
			case 't':
				timeout = atoi(optarg);
				break;
			case 'o':
				ok_sig = signum(optarg);
				break;
			case 'r':
				rpl_sig = signum(optarg);
				break;
			case 'n':
				nok_sig = signum(optarg);
				break;
			case 'N':
				nok_sig = 0;
				break;
			case 'k':
				kill_sig = signum(optarg);
				break;
			case 'K':
				kill_sig = 0;
				break;
			case 'u':
				use_ec = 1;
				break;
			case 'A':
				use_alarm = 0;
				break;
			case 'h':
				print_usage(argv[0], 1);
				return EXIT_SUCCESS;
			case 'v':
				verbose = 1;
				break;
			case 'V':
				print_version();
				return EXIT_SUCCESS;
			case '?':
				return EXIT_FAILURE;
			default:
				printf("?? getopt returned %o ??\n", c);
				return EXIT_FAILURE;
		}
	}

	if (argc - optind == 0) {
		print_usage(argv[0], 0);
		return EXIT_FAILURE;
	}

	if (timeout < 1) {
		fprintf(stderr, "timeout must be greater than one second!\n");
		return EXIT_FAILURE;
	}

	/* install signal handlers */
	signal(ok_sig, ok_sh);
	if (nok_sig != 0) {
		signal(nok_sig, nok_sh);
	}

	pid = fork();
	if (pid == 0) {
		int i;
		/* replace argv with parameters */
		for (i = optind; i < argc; i++) {
			argv[i - optind] = argv[i];
		}
		argv[i - optind] = NULL;
		argc -= optind;

		/* replace signal and */
		signal(rpl_sig, SIG_IGN);

		/* start */
		execv(argv[0], argv); /* start x */

		/* print error */
		if (verbose == 1) {
			fprintf(stderr, "failed to start: ");
			print_child_cmdline(argc, argv, 0);
			fprintf(stderr, "\n");
		}

		/* notify parent */
		if (nok_sig != 0) {
			pid = getppid();
			kill(pid, nok_sig);
		}

		/* we exit, but parent will wait -- unless it caught nok_sig */
		exit(EXIT_FAILURE);
	}

	/*
	 * Wait timeout seconds, or until child exists, whatever comes first.
	 * sleep(1) won't delay OK result, we get a signal for that. sleep(1)
	 * add unnecessary one second delay only if child returns before
	 * timeout. However if starting the child fails, we return immediately.
	 */
	while (timeout-- > 0) {
		int status;
		sleep(1);
		waitpid(pid, &status, WNOHANG);
		if (WIFEXITED(status)) {
			if (verbose == 1) {
				print_child_cmdline(argc, argv, optind);
				fprintf(stderr, " exited with code %d\n",
						WEXITSTATUS(status));
			}
			if (use_ec == 1) {
				fprintf(stderr, "using return code\n");
				return WEXITSTATUS(status);
			}
			return 1;
		}
	}

	if (verbose == 1) {
		print_child_cmdline(argc, argv, optind);
		fprintf(stderr, " didn't send signal in time\n");
	}

	/* kill timed out child */
	if (kill_sig != 0) {
		if (verbose == 1) {
			fprintf(stderr, "killing with signal %d\n", kill_sig);
		}
		/*
		 * if the process has died between waitpid() and getting here,
		 * well, adding another waitpid here is /very/ unlikely to
		 * solve anything 
		 */
		kill(pid, kill_sig);
	}

	return EXIT_FAILURE;
}
